////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// File:        dvr_protocol.h
// Function:    Public definitions, Sensoray DVR communication protocol
// Author:      Jim Lamberson
// Copyright:   Copyright (c) Sensoray Co., Inc 2020
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#ifndef _INC_DVR_PROTOCOL_H_
#define _INC_DVR_PROTOCOL_H_

#define MAX_CHUNKSIZE               240     // Maximum size of image chunk when uploading bitmap image to overlay region

// Error codes ----------------------------------------------

#define DVR_ERR_OK                  0x00    // Command executed normally
// DVR-generated:
#define DVR_ERR_INVALID_OPCODE      0x01    // Command not supported
#define DVR_ERR_ILLEGAL_ARGVAL      0x02    // Command argument value outside legal range
#define DVR_ERR_MISSING_ARG         0x03    // Missing command argument
#define DVR_ERR_FUNC_PROHIBITED     0x04    // Command function unavailable in current DVR state
#define DVR_ERR_FILE_NOT_FOUND      0x05    // Specified file not found
// DVR/API
#define DVR_ERR_MALLOC              0x06    // Can't allocate memory
// API and transaction:
#define DVR_ERR_OPEN                0x80    // Can't open DVR communication interface
#define DVR_ERR_CLOSE               0x81    // Can't close DVR communication interface
#define DVR_ERR_SOCKCREATE          0x82    // Can't create socket
#define DVR_ERR_SOCKBIND            0x83    // Socket bind failed
#define DVR_ERR_RXTHREAD            0x84    // Failed to start receiver thread
#define DVR_ERR_TXTHREAD            0x85    // Failed to start transmitter thread
#define DVR_ERR_MUTEXCREATE         0x86    // Can't create mutex
#define DVR_ERR_LOCKCLOSED          0x87    // Lock closed or timeout
#define DVR_ERR_SOCKCLOSED          0x88    // Socket closed or shutdown
#define DVR_ERR_OVERSIZE            0x90    // Command arguments too large
#define DVR_ERR_PACKETSEND          0x91    // Failed to send command to DVR
#define DVR_ERR_PACKETREAD          0xA0    // Socket closed or read error
#define DVR_ERR_RUNTPACKET          0xA1    // Abnormally short DVR reply packet
#define DVR_ERR_MISSINGDELIM        0xA2    // Abnormally long DVR reply packet
#define DVR_ERR_CHECKSUM            0xA3    // Receive checksum error
#define DVR_ERR_MSGTYPE             0xA4    // Unexpected/unknown TYPE code in DVR reply or event
#define DVR_ERR_DEPLETED            0xA5    // Missing expected data in DVR reply
#define DVR_ERR_STRINGBUF           0xA6    // String buffer too small to receive DVR reply string
#define DVR_ERR_TIMEOUT             0xA7    // Read timeout
#define DVR_ERR_EVENTTYPE           0xA8    // Unrecognized event type
#define DVR_ERR_STORAGE             0xA9    // Out of storage
#define DVR_ERR_VIDLOCK             0xAA    // Video not locked
// Menu system
#define DVR_ERR_MENUPANE_NOTFOUND   0xF0    // Menu pane doeesn't exist
#define DVR_ERR_MENUITEM_NOTFOUND   0xF1    // Menu item doesn't exist
#define DVR_ERR_MENUALREADYOPEN     0xF2    // Attempted to open menu system while already open
#define DVR_ERR_MENUCLOSED          0xF3    // Menu system is not open

// Event type codes ---------------------------------------------

#define DVR_EVENT_STATUS_CHANGE     0x01    // DVR status changed; second payload value contains bitmask identifying changed status bits

// Event client types --------------------------------------------

#define DVR_EVENTCLIENT_REMOTE      0       // Remote DVR controller
#define DVR_EVENTCLIENT_KEYPAD      1       // Local keypad
#define DVR_EVENTCLIENT_KEYBOARD    2       // Local keyboard
#define DVR_EVENTCLIENT_STATUSMON   3       // Local status monitor

// DVR status flags (u32_stat) --------------------

#define DVR_STAT_RECORDING          (1 <<  0)   // Recording; unsafe to remove memory stick
#define DVR_STAT_PLAYING            (1 <<  1)   // Video playing
#define DVR_STAT_VIDEO_LOCK         (1 <<  2)   // Input video detected
#define DVR_STAT_KBD_ATTACHED       (1 <<  3)   // Keyboard detected
#define DVR_STAT_MEM_ATTACHED       (1 <<  4)   // Memory stick mounted
#define DVR_STAT_WIFI_ENABLED       (1 <<  5)   // Wifi enabled
#define DVR_STAT_RECORD_PAUSED      (1 <<  6)   // Recording paused
#define DVR_STAT_PLAY_PAUSED        (1 <<  7)   // Playing paused
#define DVR_STAT_MENU_OPEN          (1 <<  8)   // Menu is open
#define DVR_STAT_VIEW_SNAPSHOT      (1 <<  9)   // Snapshot viewing
// Error indicators:
#define DVR_STAT_WRITE_ERROR        (1 << 16)   // Storage device full or write error
// Event notifications:
#define DVR_STAT_EVENTS_ENABLED     (1 << 24)   // Async event reporting enabled for remote client; note: event message sent upon enable but not disable
#define DVR_STAT_SERVER_SHUTDOWN    (1 << 25)   // DVR is preparing to shut down

// Command opcodes -----------------------------------

// General
#define CMD_GET_API_VERSION         1      // rx:none  tx:u32 version
#define CMD_GET_HARDWARE_INFO       2      // rx:none  tx:u32 hw_version, string modelnum
#define CMD_GET_DVR_STATUS          3      // rx:none  tx:u32 statusflags
#define CMD_ENABLE_EVENT_REPORTING  4      // rx:u8 clientid, u32 eventflags, u16 port  tx:none
#define CMD_SET_DISPLAY_MODE        5      // rx:u8 mode  tx:none
						// hd stream (first resizer), sd stream (second resizer), idle (black), colorbars
#define CMD_GET_GENERAL_SETTINGS    9      // rx:none  tx:u8 mode
						// event_enables, display_mode

// Date/time u32_clock = Epoch (seconds elapsed since 1970-01-01 00:00:00 GMT)
#define CMD_SET_DATETIME            10      // rx:u32 datetime  tx:none
#define CMD_GET_DATETIME            11      // rx:none  tx:u32 datetime

// Video record
#define CMD_RECORD_START            20      // rx:u8 devflags,u8 format,u16 bitrate_kbps tx:string filename
#define CMD_RECORD_STOP             21      // rx:none  tx:none
#define CMD_RECORD_PAUSE            22      // rx:none  tx:none
#define CMD_RECORD_RESUME           23      // rx:none  tx:none

// Video play
#define CMD_PLAY_SKIP               31      // rx:s16 seconds  tx:none
#define CMD_PLAY_START              32      // rx:u8 dev, string filename  tx:none
#define CMD_PLAY_STOP               33      // rx:none  tx:none
#define CMD_PLAY_PAUSE              34      // rx:none  tx:none
#define CMD_PLAY_RESUME             35      // rx:none  tx:none
#define CMD_SET_PLAY_SPEED          36      // rx:u8 speed  tx:none
#define CMD_GET_PLAY_SETTINGS       39      // rx:none  tx:u8 speed

// Snapshot
#define CMD_VIEW_SNAPSHOT           40      // rx:u8 dev, string filename  tx:none
#define CMD_GRAB_SNAPSHOT           41      // rx:u8 dev, u8 quality  tx:string filename
#define CMD_SNAPSHOT_STOP           43      // rx:none  tx:none

// File system  u8_devid = DVR_STORAGE_INTERNAL | DVR_STORAGE_EXTERNAL
#define CMD_GET_FILECOUNT           50      // rx:u8 dev  tx:u32 count
#define CMD_GET_FILENAME            52      // rx:u8 dev,u32 index  tx:string filename
#define CMD_DELETE_FILE             53      // rx:u8 dev,string filename  tx:none
#define CMD_RENAME_FILE             54      // rx:u8 dev,string oldname,string newname  tx:none

// Video overlay  u8_region = overlay region ID in range 0:3
#define CMD_SET_OSD_TEXT            60      // rx:u8 region,string text  tx:none
#define CMD_SET_OSD_POSITION        61      // rx:u8 region,u16 top,u16 left  tx:none
#define CMD_SET_OSD_ENABLE          62      // rx:u8 region,u8 enable  tx:none
#define CMD_SET_OSD_BACKGROUND      63      // rx:u8 region,u8 bgmode  tx:none
#define CMD_SET_OSD_CLOCK_FORMAT    64      // rx:u8 datefmt, u8 timefmt  tx:none
#define CMD_GET_OSD_CLOCK_FORMAT    65      // rx:none  tx:u8 datefmt, u8 timefmt
#define CMD_GET_OSD_SETTINGS        69      // rx:u8 region  tx:u8 enable, u8 bgmode, u16 top, u16 left
#define CMD_OSD_IMAGE_CREATE        70      // rx:u8 region,u16 len  tx:none
#define CMD_OSD_IMAGE_UPLOAD_CHUNK  71      // rx:u8 region,u16 addr,u8[] payload  tx:none
#define CMD_OSD_IMAGE_INSTALL       72      // rx:u8 region,u8 imagetype  tx:

// Video adjust   u8_paramid = parameter ID (see DVR_VIDEO_ADJ_ID); u8_setting = parameter value
#define CMD_SET_VIDEO_ADJUST        80      // rx:u8 param_id, u8 value  tx:none
#define CMD_GET_VIDEO_ADJUST        81      // rx:u8 param_id  tx:u8 value

// Video configuration   u8_format = see DVR_VIDEO_FORMAT; u8_deint = deinterlace enable (1 | 0); u8_decimate = frame rate divisor in range 1:input_framerate (1 = full frame rate)
#define CMD_SET_VIDEO_CFG           100     // rx:u8 format, u8 deint, u8 decimate  tx:none
#define CMD_GET_VIDEO_CFG           101     // rx:none  tx:u8 format, u8 deint, u8 decimate

// Audio  u8_chansel = AUDIOCHAN_LEFT | AUDIOCHAN_RIGHT | AUDIOCHAN_BOTH; 
#define CMD_SET_AUDIO_INPUT         110     // rx:u8 src  tx:none
#define CMD_SET_VOLUME              111     // rx:u8 chansel,u8 volume  tx:none
#define CMD_SET_MUTE                112     // rx:u8 chansel,u8 mute  tx:none
#define CMD_SET_AUTOVOL             113     // rx:u8 chansel,u8 agc  tx:none
#define CMD_GET_AUDIO_SETTINGS      119     // rx:none  tx:u8 src, u8 lvol, u8 rvol, u8 lmute, u8 rmute, u8 lagc, u8 ragc

// Incremental encoder  u8_chan = encoder channel number in range 0:num_encoders; u32_counts = 32-bit position counts
#define CMD_SET_ENCODER_COUNTS      120     // rx:u8 chan,u32 count  tx:none
#define CMD_GET_ENCODER_COUNTS      121     // rx:u8 chan  tx:u32 count
#define CMD_SET_ENCODER_SCALAR      122     // rx:u8 chan,double scale  tx:none
#define CMD_USE_OFFBOARD_ENCODER    123     // rx:u8 chan,u8 enable  tx:none
#define CMD_GET_ENCODER_SETTINGS    129     // rx:u8 chan  tx:double scalar, u8 use_offboard

// Menu system
#define CMD_MENU_DISPLAY            140     // rx:u8_enable  tx:none
#define CMD_MENU_MOVE_CURSOR        141     // rx:u8_up  tx:none
#define CMD_MENU_ENTER              142     // rx:none  tx:u8_menuid, u8_executable, string_menutext
#define CMD_MENU_EXIT               143     // rx:none  tx:none

// Message type codes -------------------------------

#define MSGTYPE_COMMAND             0x00
#define MSGTYPE_REPLY               0x40
#define MSGTYPE_EVENT               0x80

// Packet attributes -------------------

#define SIZEOF_CHECKSUM             1
#define SIZEOF_STRING_TERMINATOR    1
#define SIZEOF_COBSHDR              1
#define SIZEOF_PACKET_DELIM         1
#define MIN_PACKET_SIZE             4       // Minimum packet size: cobshdr, msgtype, opcode/evtype/errcode, checksum
#define MAX_PACKET_SIZE             255
#define PACKET_DELIMITER            0

#define MAX_FILENAME_SIZE       120

// Types --------------------------------------------

// Storage devices
#define DVR_STORAGE_INTERNAL        (1 << 0)
#define DVR_STORAGE_EXTERNAL        (1 << 1)
typedef enum DVR_STORAGE_DEVID {
    MMC         = 0,
    INTERNAL    = DVR_STORAGE_INTERNAL,
    EXTERNAL    = DVR_STORAGE_EXTERNAL,
    BOTH        = DVR_STORAGE_INTERNAL + DVR_STORAGE_EXTERNAL
} DVR_STORAGE_DEVID;

typedef enum DVR_OSD_IMAGETYPE {
    PNG = 0,
    BMP = 1
} DVR_OSD_IMAGETYPE;

typedef enum DVR_OSD_BGMODE {
    BGMODE_BLACK,
    BGMODE_TRANSPARENT
} DVR_OSD_BGMODE;

typedef enum DVR_VIDEO_ADJ_ID {    // Video adjust ID
    ADJ_BRIGHTNESS,
    ADJ_AUTOBRIGHT,
    ADJ_CONTRAST,
    ADJ_HUE,
    ADJ_SATURATION,
    ADJ_CHROMA_FILT,
    ADJ_LUMA_FILT,
    ADJ_COMB_FILT
} DVR_VIDEO_ADJ_ID;

typedef enum DVR_VIDEO_FORMAT {
    VIDFMT_NTSC,
    VIDFMT_PAL
        // tbd: HD formats
} DVR_VIDEO_FORMAT;

typedef enum DVR_VIDREC_FORMAT {
    RECFMT_MP4  = 0,
    RECFMT_TS   = 1
} DVR_VIDREC_FORMAT;

typedef enum DVR_DATE_FORMAT {
    M2_D2_Y4,       // MM-DD-YYYY
    D2_M2_Y4,       // DD-MM-YYYY
    M2_D2_Y2,       // MM-DD-YY
    D2_M2_Y2,       // DD-MM-YY
    M3_D2_Y4,       // MMM-DD-YYYY
    D2_M3_Y4,       // DD-MMM-YYYY
    M3_D2_Y2,       // MMM-DD-YY
    D2_M3_Y2,       // DD-MMM-YY
    M3_D2_Y4_STR,   // MMM DD, YYYY
    D2M3Y4,         // DDMMMYYYY
    M3_D2_Y2_STR,   // MMM DD, YY
    D2M3Y2,         // DDMMMYY
    Y4M2D2,         // YYYYMMDD
    Y4M3D2,         // YYYYMMMDD
    MAX_DATE_FORMAT
} DVR_DATE_FORMAT;

typedef enum DVR_TIME_FORMAT {
    HH_MM_SS,       //"HH:MM:SS", 
    HH_MM_SS_F1,    //"HH:MM:SS.S", 
    HH_MM_SS_F2,    //"HH:MM:SS.SS",
    HH_MM_SS_F3,    //"HH:MM:SS.SSS", 
    HH_MM,          //"HH:MM", 
    HH_MM_AP,       //"HH:MM AM/PM", 
    HH_MM_ap,       //"HH:MM a/p"
    MAX_TIME_FORMAT
} DVR_TIME_FORMAT;

typedef enum DVR_DISPLAY_MODE {
    DISPMODE_HD_STREAM,     // hd stream (first resizer)
    DISPMODE_SD_STREAM,     // sd stream (second resizer)
    DISPMODE_BLACK,         // idle (black)
    DISPMODE_COLORBAR       // colorbars
} DVR_DISPLAY_MODE;

typedef enum DVR_AUDIO_CHAN {
    AUDIOCHAN_LEFT  = 1,
    AUDIOCHAN_RIGHT = 2,
    AUDIOCHAN_BOTH  = 3
} DVR_AUDIO_CHAN;

typedef enum DVR_AUDIO_INPUT {
    AUDIOIN_MIC,
    AUDIOIN_LINE,
    AUDIOIN_DIGITAL
} DVR_AUDIO_INPUT;
    
typedef enum DVR_PLAY_SPEED {
    PLAY_NORMAL,
    PLAY_FAST_FORWARD,
    PLAY_FAST_REVERSE,
} DVR_PLAY_SPEED;



#endif  // #ifndef _INC_DVR_PROTOCOL_H_
