////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// File:        simple_sdvr_codec.h
// Function:    Sensoray DVR communication protocol, example utility functions and usage
// Copyright:   Copyright (c) Sensoray Co., Inc 2020
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include <time.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>

#include "dvr_protocol.h"

typedef unsigned char   u8;     // 8-bit unsigned
typedef unsigned short  u16;    // 16-bit unsigned
typedef   signed short  s16;    // 16-bit signed
typedef unsigned int    u32;    // 32-bit unsigned


//////////////////////////////////////////////////////////////
// User-supplied serial communication functions

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <termios.h>
#include <poll.h>

static int com_fd = -1;
static struct termios saved_tty;

// recv_byte will wait 1s before timeout
static int timeout_ms = 1000;

int open_com_port(void)
{
    char *device = "/dev/ttyACM0";
    int baudrate = 115200;
    struct termios tty;
    speed_t speed = 0;
    
    if (com_fd != -1) {
        // already open
        return DVR_ERR_OK;
    }

    com_fd = open(device, O_RDWR);
    if (com_fd == -1) {
	if (errno == ENOENT) {
	    device = "/dev/ttyACM1";
	    com_fd = open(device, O_RDWR);
	}
	if (com_fd == -1) {
            perror("open com port");
            return DVR_ERR_OPEN;
        }
    }
    
    printf("%s %d\n", device, baudrate);

    // convert numerical baudrate to baudrate define
    switch (baudrate) {
        case 50: speed = B50; break;
        case 75: speed = B75; break;
        case 110: speed = B110; break;
        case 134: speed = B134; break;
        case 150: speed = B150; break;
        case 200: speed = B200; break;
        case 300: speed = B300; break;
        case 600: speed = B600; break;
        case 1200: speed = B1200; break;
        case 1800: speed = B1800; break;
        case 2400: speed = B2400; break;
        case 4800: speed = B4800; break;
        case 9600: speed = B9600; break;
        case 19200: speed = B19200; break;
        case 38400: speed = B38400; break;
        case 57600: speed = B57600; break;
        case 115200: speed = B115200; break;
        case 230400: speed = B230400; break;
        case 460800: speed = B460800; break;
        case 500000: speed = B500000; break;
        case 576000: speed = B576000; break;
        case 921000: speed = B921600; break;
        case 1000000: speed = B1000000; break;
        case 1152000: speed = B1152000; break;
        case 1500000: speed = B1500000; break;
        case 2000000: speed = B2000000; break;
        case 2500000: speed = B2500000; break;
        case 3000000: speed = B3000000; break;
        case 3500000: speed = B3500000; break;
        case 4000000: speed = B4000000; break;
        default: return DVR_ERR_OPEN;
    }

    // Set speed to 115,200 bps, 8n1 (no parity)
    memset (&tty, 0, sizeof tty);
    if (tcgetattr (com_fd, &tty) != 0) {
        perror("tcgetattr");
        close(com_fd);
        com_fd = -1;
        return DVR_ERR_OPEN;
    }
    saved_tty = tty;

    //printf("baudrate=%d speed=%d\n", baudrate, speed);

    if (cfsetospeed(&tty, speed) < 0)
        perror("cfsetospeed");
    if (cfsetispeed(&tty, speed) < 0)
        perror("cfsetispeed");

    tty.c_cflag     = (tty.c_cflag & ~CSIZE) | CS8; // 8 data bits
    tty.c_iflag     &= ~IGNBRK;                     // ignore breaks
    tty.c_iflag     &= ~(INLCR | ICRNL);            // don't translate NL to CR on input and vice versa
    tty.c_lflag     = 0;                            // non-canonical (raw) mode, no signaling chars, no echo
    tty.c_oflag     = 0;                            // no remapping, no delays
    tty.c_cc[VMIN]  = 1;                            // blocking read
    tty.c_cc[VTIME] = 1;                            // 100 milliseconds read timeout
    tty.c_iflag     &= ~(IXON | IXOFF | IXANY);     // disable xon/xoff control
    tty.c_cflag     |= (CLOCAL | CREAD);            // ignore modem controls
    tty.c_cflag     &= ~(PARENB | PARODD);          // disable parity
    tty.c_cflag     &= ~CSTOPB;                     // one stop bit
    tty.c_cflag     &= ~CRTSCTS;                    // disable hardware flow control

    tcflush(com_fd, TCIFLUSH);
    if (tcsetattr(com_fd, TCSANOW, &tty) != 0) {
        perror("tcsetattr TCSANOW");
        close(com_fd);
        com_fd = -1;
        return DVR_ERR_OPEN;
    }
    
    return DVR_ERR_OK;
}

void close_com_port(void)
{
    if (com_fd != -1) {
        if (tcsetattr(com_fd, TCSANOW, &saved_tty) != 0)
            perror("tcsetattr TCSANOW");
        close(com_fd);
        com_fd = -1;
    }
}

static int send_byte(u8 byte)
{
    // TODO : SEND BYTE TO DVR VIA SERIAL TRANSMITTER. RETURNS: DVR_ERR_OK or DVR_ERR_PACKETSEND
    if (write(com_fd, &byte, 1) < 0)
        perror("send_byte.write");
    return DVR_ERR_OK;
}

int send_bytes(u8 *bytes, int len)
{
    // TODO : SEND BYTE TO DVR VIA SERIAL TRANSMITTER. RETURNS: DVR_ERR_OK or DVR_ERR_PACKETSEND
    if (write(com_fd, bytes, len) < 0)
        perror("send_byte.write");
    return DVR_ERR_OK;
}

static int recv_byte(u8 *byte)
{
    // TODO : RECEIVE BYTE FROM DVR VIA SERIAL RECEIVER. RETURNS: DVR_ERR_OK or DVR_ERR_PACKETREAD
    struct pollfd pfd = { .fd = com_fd, .events = POLLIN };
    int rc;
    
    while ((rc = poll(&pfd, 1, timeout_ms)) <= 0) {
        if (rc < 0) {
            perror("poll");
            return DVR_ERR_PACKETREAD;
        }
	return DVR_ERR_TIMEOUT;
    }

    if (read(com_fd, byte, 1) < 0) {
        perror("recv_byte.read");
        return DVR_ERR_PACKETREAD;
    }
    return DVR_ERR_OK;
}



/////////////////////////////////////////////////////////////////////////////////////
// In-place DVR packet encoder.
// Note: packet buffer (pktbuf) must have one byte of additional space preceeding
// TYPE for cobs header byte, and two bytes following BODY for SUM and packet delimiter.
// Imports:
//  pktbuf - unencoded TYPE and BODY starting at pktbuf[1]
//  nbytes - size of TYPE and BODY
// Exports:
//  pktbuf - cobs-encoded packet

static void dvr_encode(u8 *pktbuf, u8 nbytes)
{
    u8 val = 0; // checksum/stuffval
    u8 *pbuf = pktbuf + 1;

    while (nbytes--)                        // compute checksum
        val += *pbuf++;
    *pbuf++ = 0 - val;                      // append checksum
    *pbuf-- = PACKET_DELIMITER;             // append packet delimiter
    for (val = 1; pbuf > pktbuf; pbuf--) {  // reverse cobs-encode starting at SUM
        if (*pbuf == 0) {
            *pbuf = val; // stuff
            val = 1;
        } else
            val++;
    }
    *pbuf = val;
}

static u8 *packet_end = NULL;

/////////////////////////////////////////////////////////////////////////////////////
// In-place DVR packet decoder.
// Imports:
//  pktbuf - cobs-encoded packet (including packet delimiter 0x00 at end)
//  buflen - size of packet buffer
// Exports:
//  pktbuf - decoded TYPE and BODY starting at pktbuf[1]
// Return value:
//  error code

static int dvr_decode(u8 *pktbuf, u8 buflen)
{
    u8 *pend = pktbuf + buflen;
    u8 *pbuf = pktbuf + pktbuf[0];              // find first stuff position
    
    while (pbuf < pend) {                       // repeat while position is within pktbuf:
        if (*pbuf == PACKET_DELIMITER) {                    // if end of packet
            packet_end = pbuf;
            u8 sum = *--pbuf;                               //   init checksum = SUM
            u8 datalen = --pbuf - pktbuf;                   //   compute data size
            while (pbuf > pktbuf)                           //   compute checksum
                sum += *pbuf--;
            return sum ? DVR_ERR_CHECKSUM : DVR_ERR_OK;     //   validate checksum and exit
        } else {                                            // else cobs-decode:
            u8 *pzero = pbuf;                               //   cache current position
            pbuf += *pbuf;                                  //   find next stuff position
            *pzero = 0;                                     //   unstuff current
        }
    }
    return DVR_ERR_MISSINGDELIM;  // decoding error: missing packet delimiter
}

///////////////////////////////////////////////////
// Send packet to DVR via serial transmitter.

static int send_packet(u8 *buf)
{
    int errcode;
    do {
        errcode = send_byte(*buf);
    }
    while (*buf++ && !errcode);  // send until packet delimiter sent or transmitter error
    return errcode;
}

////////////////////////////////////////////////////
// Receive packet from DVR via serial receiver.



static void print_packet(u8 *pktbuf);  // TEMPORARY SIMULATION



static int recv_packet(u8 *buf, u8 buflen)
{

    int errcode = DVR_ERR_OK;
    do {
        errcode = recv_byte(buf);
    }
    while (--buflen && *buf++ && !errcode);  // receive until buffer full, packet delimiter received, or receiver error
    return errcode;
}


////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////  EXAMPLE USAGE  //////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////


#define DEBUG   // uncomment to print packets

// Utility function for debug only: show packet bytes in hex
static void print_packet(u8 *pktbuf)
{
    int i;
    for (i = 0; pktbuf[i]; i++)
        printf ("%02X ", pktbuf[i]);
    printf ("00\n");
}

// Fetch 32-bit value from dvr reply
static void extract_u32(u8 *buf, u32 *dest)
{
    if (buf + 4 > packet_end) {
        printf("%s: read past end\n", __func__);
    }
    *dest = (((((buf[3] << 8) + buf[2]) << 8) + buf[1]) << 8) + buf[0];
}

// Fetch double value from dvr reply
static void extract_double(u8 *buf, double *dest)
{
    if (buf + 8 > packet_end) {
        printf("%s: read past end\n", __func__);
    }
    memcpy(dest, buf, sizeof(double));
}

// Fetch string and terminator from dvr reply and return the number of bytes
static int extract_string(u8 *buf, char *dest)
{
    int len = strlen((char*)buf) + SIZEOF_STRING_TERMINATOR;
    if (buf + len + 1 > packet_end) {
        printf("%s: read past end\n", __func__);
    }
    memcpy(dest, buf, len);
    return len;
}


// Send command to dvr and receive/validate reply
static int dvr_transact(u8 *pktbuf, u8 buflen, u8 datalen)
{
    int errcode;
    dvr_encode(pktbuf, datalen);                                                        // cobs-encode and packetize
#ifdef DEBUG
    printf("tx: "); print_packet(pktbuf);
#endif
    if ((errcode = send_packet(pktbuf)) == DVR_ERR_OK)                                  // send command packet to dvr
        if ((errcode = recv_packet(pktbuf, buflen)) == DVR_ERR_OK) {                     // receive reply packet
#ifdef DEBUG
	    printf("rx: "); print_packet(pktbuf);
#endif
            if ((errcode = dvr_decode(pktbuf, buflen)) == DVR_ERR_OK)                   // decode reply
                errcode = (pktbuf[1] == MSGTYPE_REPLY) ? pktbuf[2] : DVR_ERR_MSGTYPE;   // return dvr-generated errcode (or DVR_ERR_MSGTYPE if packet isn't a reply)
        }
#ifdef DEBUG
    if (errcode != DVR_ERR_OK) {
        printf("%s: errcode=%x\n", __func__, errcode);
    }
#endif
    return errcode;
}



////////////////////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////////  SIMPLE COMMANDS /////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////

static u32 last_status = 0;

u32 dvr_get_last_status(void)
{
    return last_status;
}

int dvr_get_api_version(u32 *version)
{
    // Build command packet
    int errcode;
    u8 pktbuf[100]; // send/recv packet buf
    u8 *pbuf = &pktbuf[1];                  // reserve space for cobs header
                                            // build command:
    *pbuf++ = MSGTYPE_COMMAND;              //   message type
    *pbuf++ = CMD_GET_API_VERSION;          //   opcode

    // Send command and receive reply
    errcode = dvr_transact(pktbuf, sizeof(pktbuf), pbuf - &pktbuf[1]);
    if (errcode == DVR_ERR_OK) {
        extract_u32(pktbuf + 3, version);
        extract_u32(pktbuf + 7, &last_status);
    }
    return errcode;
}

int dvr_get_hardware_info(u32 *hw_version, char *model)
{
    // Build command packet
    int errcode;
    u8 pktbuf[100]; // send/recv packet buf
    u8 *pbuf = &pktbuf[1];                  // reserve space for cobs header
    int textlen;
                                            // build command:
    *pbuf++ = MSGTYPE_COMMAND;              //   message type
    *pbuf++ = CMD_GET_HARDWARE_INFO;        //   opcode

    // Send command and receive reply
    errcode = dvr_transact(pktbuf, sizeof(pktbuf), pbuf - &pktbuf[1]);
    if (errcode == DVR_ERR_OK) {
        extract_u32(pktbuf + 3, hw_version);
        textlen = extract_string(pktbuf + 7, model);
        extract_u32(pktbuf + 7 + textlen, &last_status);
    }
    return errcode;
}

int dvr_get_dvr_status(u32 *status)
{
    // Build command packet
    int errcode;
    u8 pktbuf[100]; // send/recv packet buf
    u8 *pbuf = &pktbuf[1];                  // reserve space for cobs header
                                            // build command:
    *pbuf++ = MSGTYPE_COMMAND;              //   message type
    *pbuf++ = CMD_GET_DVR_STATUS;           //   opcode

    // Send command and receive reply
    errcode = dvr_transact(pktbuf, sizeof(pktbuf), pbuf - &pktbuf[1]);
    if (errcode == DVR_ERR_OK) {
        extract_u32(pktbuf + 3, status);
    }
    return errcode;
}

int dvr_enable_event_reporting(u8 clientid, u32 eventflags, u16 port)
{
    // Build command packet
    int errcode;
    u8 pktbuf[100]; // send/recv packet buf
    u8 *pbuf = &pktbuf[1];                  // reserve space for cobs header
                                            // build command:
    *pbuf++ = MSGTYPE_COMMAND;              //   message type
    *pbuf++ = CMD_ENABLE_EVENT_REPORTING;   //   opcode
    *pbuf++ = clientid;
    *pbuf++ = (u8)eventflags;
    *pbuf++ = (u8)(eventflags>>8);
    *pbuf++ = (u8)(eventflags>>16);
    *pbuf++ = (u8)(eventflags>>24);
    *pbuf++ = (u8)port;
    *pbuf++ = (u8)(port>>8);

    // Send command and receive reply
    errcode = dvr_transact(pktbuf, sizeof(pktbuf), pbuf - &pktbuf[1]);
    if (errcode == DVR_ERR_OK) {
        extract_u32(pktbuf + 3, &last_status);
    }
    return errcode;
}

int dvr_set_display_mode(DVR_DISPLAY_MODE mode)
{
    // Build command packet
    int errcode;
    u8 pktbuf[100]; // send/recv packet buf
    u8 *pbuf = &pktbuf[1];                  // reserve space for cobs header
                                            // build command:
    *pbuf++ = MSGTYPE_COMMAND;              //   message type
    *pbuf++ = CMD_SET_DISPLAY_MODE;         //   opcode
    *pbuf++ = (u8)mode;

    // Send command and receive reply
    errcode = dvr_transact(pktbuf, sizeof(pktbuf), pbuf - &pktbuf[1]);
    if (errcode == DVR_ERR_OK) {
        extract_u32(pktbuf + 3, &last_status);
    }
    return errcode;
}

int dvr_get_general_settings(DVR_DISPLAY_MODE *mode)
{
    // Build command packet
    int errcode;
    u8 pktbuf[100]; // send/recv packet buf
    u8 *pbuf = &pktbuf[1];                  // reserve space for cobs header
                                            // build command:
    *pbuf++ = MSGTYPE_COMMAND;              //   message type
    *pbuf++ = CMD_GET_GENERAL_SETTINGS;     //   opcode

    // Send command and receive reply
    errcode = dvr_transact(pktbuf, sizeof(pktbuf), pbuf - &pktbuf[1]);
    if (errcode == DVR_ERR_OK) {
        *mode = (DVR_DISPLAY_MODE)pktbuf[3];
        extract_u32(pktbuf + 4, &last_status);
    }
    return errcode;
}


// Store the date and time into the DVR, clock = Epoch (seconds elapsed since 1970-01-01 00:00:00 GMT)
int dvr_set_datetime(u32 datetime)
{
    // Build command packet
    int errcode;
    u8 pktbuf[100]; // send/recv packet buf
    u8 *pbuf = &pktbuf[1];                  // reserve space for cobs header
                                            // build command:
    *pbuf++ = MSGTYPE_COMMAND;              //   message type
    *pbuf++ = CMD_SET_DATETIME;             //   opcode
    *pbuf++ = (u8)datetime;
    *pbuf++ = (u8)(datetime>>8);
    *pbuf++ = (u8)(datetime>>16);
    *pbuf++ = (u8)(datetime>>24);

    // Send command and receive reply
    errcode = dvr_transact(pktbuf, sizeof(pktbuf), pbuf - &pktbuf[1]);
    if (errcode == DVR_ERR_OK) {
        extract_u32(pktbuf + 3, &last_status);
    }
    return errcode;
}

// Read the date and time from the DVR, clock = Epoch (seconds elapsed since 1970-01-01 00:00:00 GMT)
int dvr_get_datetime(u32 *datetime)
{
    // Build command packet
    int errcode;
    u8 pktbuf[100]; // send/recv packet buf
    u8 *pbuf = &pktbuf[1];                  // reserve space for cobs header
                                            // build command:
    *pbuf++ = MSGTYPE_COMMAND;              //   message type
    *pbuf++ = CMD_GET_DATETIME;             //   opcode

    // Send command and receive reply
    errcode = dvr_transact(pktbuf, sizeof(pktbuf), pbuf - &pktbuf[1]);
    if (errcode == DVR_ERR_OK) {
        extract_u32(pktbuf + 3, datetime);
        extract_u32(pktbuf + 7, &last_status);
    }
    return errcode;
}


int dvr_record_start(DVR_STORAGE_DEVID devflags, DVR_VIDREC_FORMAT format, u16 bitrate, char *filename)
{
    // Build command packet
    int errcode;
    u8 pktbuf[100]; // send/recv packet buf
    u8 *pbuf = &pktbuf[1];                  // reserve space for cobs header
    int textlen;
                                            // build command:
    *pbuf++ = MSGTYPE_COMMAND;              //   message type
    *pbuf++ = CMD_RECORD_START;             //   opcode
    *pbuf++ = devflags;
    *pbuf++ = format;
    *pbuf++ = (u8)bitrate;
    *pbuf++ = (u8)(bitrate >> 8);

    // Send command and receive reply
    errcode = dvr_transact(pktbuf, sizeof(pktbuf), pbuf - &pktbuf[1]);
    if (errcode == DVR_ERR_OK) {
        textlen = extract_string(pktbuf + 3, filename);
        extract_u32(pktbuf + 3 + textlen, &last_status);
    }
    return errcode;
}

int dvr_record_stop(void)
{
    // Build command packet
    int errcode;
    u8 pktbuf[100]; // send/recv packet buf
    u8 *pbuf = &pktbuf[1];                  // reserve space for cobs header
                                            // build command:
    *pbuf++ = MSGTYPE_COMMAND;              //   message type
    *pbuf++ = CMD_RECORD_STOP;              //   opcode

    // Send command and receive reply
    errcode = dvr_transact(pktbuf, sizeof(pktbuf), pbuf - &pktbuf[1]);
    if (errcode == DVR_ERR_OK) {
        extract_u32(pktbuf + 3, &last_status);
    }
    return errcode;
}

int dvr_record_pause(void)
{
    // Build command packet
    int errcode;
    u8 pktbuf[100]; // send/recv packet buf
    u8 *pbuf = &pktbuf[1];                  // reserve space for cobs header
                                            // build command:
    *pbuf++ = MSGTYPE_COMMAND;              //   message type
    *pbuf++ = CMD_RECORD_PAUSE;             //   opcode

    // Send command and receive reply
    errcode = dvr_transact(pktbuf, sizeof(pktbuf), pbuf - &pktbuf[1]);
    if (errcode == DVR_ERR_OK) {
        extract_u32(pktbuf + 3, &last_status);
    }
    return errcode;
}

int dvr_record_resume(void)
{
    // Build command packet
    int errcode;
    u8 pktbuf[100]; // send/recv packet buf
    u8 *pbuf = &pktbuf[1];                  // reserve space for cobs header
                                            // build command:
    *pbuf++ = MSGTYPE_COMMAND;              //   message type
    *pbuf++ = CMD_RECORD_RESUME;            //   opcode

    // Send command and receive reply
    errcode = dvr_transact(pktbuf, sizeof(pktbuf), pbuf - &pktbuf[1]);
    if (errcode == DVR_ERR_OK) {
        extract_u32(pktbuf + 3, &last_status);
    }
    return errcode;
}


// Video Play
int dvr_play_skip(s16 seconds)
{
    // Build command packet
    int errcode;
    u8 pktbuf[100]; // send/recv packet buf
    u8 *pbuf = &pktbuf[1];                  // reserve space for cobs header
                                            // build command:
    *pbuf++ = MSGTYPE_COMMAND;              //   message type
    *pbuf++ = CMD_PLAY_SKIP;                //   opcode
    *pbuf++ = (u8)seconds;
    *pbuf++ = (u8)(seconds >> 8);

    // Send command and receive reply
    errcode = dvr_transact(pktbuf, sizeof(pktbuf), pbuf - &pktbuf[1]);
    if (errcode == DVR_ERR_OK) {
        extract_u32(pktbuf + 3, &last_status);
    }
    return errcode;
}

int dvr_play_start(DVR_STORAGE_DEVID dev, const char *filename)
{
    // Build command packet
    int errcode;
    u8 pktbuf[100]; // send/recv packet buf
    u8 *pbuf = &pktbuf[1];                  // reserve space for cobs header
    int textlen = strlen(filename) + SIZEOF_STRING_TERMINATOR;
                                            // build command:
    *pbuf++ = MSGTYPE_COMMAND;              //   message type
    *pbuf++ = CMD_PLAY_START;               //   opcode
    *pbuf++ = (u8)dev;
    memcpy(pbuf, filename, textlen);
    pbuf += textlen;

    // Send command and receive reply
    errcode = dvr_transact(pktbuf, sizeof(pktbuf), pbuf - &pktbuf[1]);
    if (errcode == DVR_ERR_OK) {
        extract_u32(pktbuf + 3, &last_status);
    }
    return errcode;
}

int dvr_play_stop(void)
{
    // Build command packet
    int errcode;
    u8 pktbuf[100]; // send/recv packet buf
    u8 *pbuf = &pktbuf[1];                  // reserve space for cobs header
                                            // build command:
    *pbuf++ = MSGTYPE_COMMAND;              //   message type
    *pbuf++ = CMD_PLAY_STOP;                //   opcode

    // Send command and receive reply
    errcode = dvr_transact(pktbuf, sizeof(pktbuf), pbuf - &pktbuf[1]);
    if (errcode == DVR_ERR_OK) {
        extract_u32(pktbuf + 3, &last_status);
    }
    return errcode;
}
        
int dvr_play_pause(void)
{
    // Build command packet
    int errcode;
    u8 pktbuf[100]; // send/recv packet buf
    u8 *pbuf = &pktbuf[1];                  // reserve space for cobs header
                                            // build command:
    *pbuf++ = MSGTYPE_COMMAND;              //   message type
    *pbuf++ = CMD_PLAY_PAUSE;               //   opcode

    // Send command and receive reply
    errcode = dvr_transact(pktbuf, sizeof(pktbuf), pbuf - &pktbuf[1]);
    if (errcode == DVR_ERR_OK) {
        extract_u32(pktbuf + 3, &last_status);
    }
    return errcode;
}
        
int dvr_play_resume(void)
{
    // Build command packet
    int errcode;
    u8 pktbuf[100]; // send/recv packet buf
    u8 *pbuf = &pktbuf[1];                  // reserve space for cobs header
                                            // build command:
    *pbuf++ = MSGTYPE_COMMAND;              //   message type
    *pbuf++ = CMD_PLAY_RESUME;              //   opcode

    // Send command and receive reply
    errcode = dvr_transact(pktbuf, sizeof(pktbuf), pbuf - &pktbuf[1]);
    if (errcode == DVR_ERR_OK) {
        extract_u32(pktbuf + 3, &last_status);
    }
    return errcode;
}

int dvr_set_play_speed(u8 speed)
{
    // Build command packet
    int errcode;
    u8 pktbuf[100]; // send/recv packet buf
    u8 *pbuf = &pktbuf[1];                  // reserve space for cobs header
                                            // build command:
    *pbuf++ = MSGTYPE_COMMAND;              //   message type
    *pbuf++ = CMD_SET_PLAY_SPEED;           //   opcode
    *pbuf++ = speed;

    // Send command and receive reply
    errcode = dvr_transact(pktbuf, sizeof(pktbuf), pbuf - &pktbuf[1]);
    if (errcode == DVR_ERR_OK) {
        extract_u32(pktbuf + 3, &last_status);
    }
    return errcode;
}

int dvr_get_play_settings(u8 *speed)
{
    // Build command packet
    int errcode;
    u8 pktbuf[100]; // send/recv packet buf
    u8 *pbuf = &pktbuf[1];                  // reserve space for cobs header
                                            // build command:
    *pbuf++ = MSGTYPE_COMMAND;              //   message type
    *pbuf++ = CMD_GET_PLAY_SETTINGS;        //   opcode

    // Send command and receive reply
    errcode = dvr_transact(pktbuf, sizeof(pktbuf), pbuf - &pktbuf[1]);
    if (errcode == DVR_ERR_OK) {
        *speed = pktbuf[3];
        extract_u32(pktbuf + 4, &last_status);
    }
    return errcode;
}

// Snapshot
int dvr_view_snapshot(DVR_STORAGE_DEVID dev, const char *filename)
{
    // Build command packet
    int errcode;
    u8 pktbuf[100]; // send/recv packet buf
    u8 *pbuf = &pktbuf[1];                  // reserve space for cobs header
    int textlen = strlen(filename) + SIZEOF_STRING_TERMINATOR;
                                            // build command:
    *pbuf++ = MSGTYPE_COMMAND;              //   message type
    *pbuf++ = CMD_VIEW_SNAPSHOT;            //   opcode
    *pbuf++ = (u8)dev;
    memcpy(pbuf, filename, textlen);
    pbuf += textlen;

    // Send command and receive reply
    errcode = dvr_transact(pktbuf, sizeof(pktbuf), pbuf - &pktbuf[1]);
    if (errcode == DVR_ERR_OK) {
        extract_u32(pktbuf + 3, &last_status);
    }
    return errcode;
}

int dvr_snapshot_stop(void)
{
    // Build command packet
    int errcode;
    u8 pktbuf[100]; // send/recv packet buf
    u8 *pbuf = &pktbuf[1];                  // reserve space for cobs header

                                            // build command:
    *pbuf++ = MSGTYPE_COMMAND;              //   message type
    *pbuf++ = CMD_SNAPSHOT_STOP;            //   opcode

    // Send command and receive reply
    errcode = dvr_transact(pktbuf, sizeof(pktbuf), pbuf - &pktbuf[1]);
    if (errcode == DVR_ERR_OK) {
        extract_u32(pktbuf + 3, &last_status);
    }
    return errcode;
}


int dvr_grab_snapshot(DVR_STORAGE_DEVID dev, u8 quality, char *filename)
{
    // Build command packet
    int errcode;
    u8 pktbuf[100]; // send/recv packet buf
    u8 *pbuf = &pktbuf[1];                  // reserve space for cobs header
    int textlen;
                                            // build command:
    *pbuf++ = MSGTYPE_COMMAND;              //   message type
    *pbuf++ = CMD_GRAB_SNAPSHOT;            //   opcode
    *pbuf++ = (u8)dev;
    *pbuf++ = quality;

    // Send command and receive reply
    errcode = dvr_transact(pktbuf, sizeof(pktbuf), pbuf - &pktbuf[1]);
    if (errcode == DVR_ERR_OK) {
        textlen = extract_string(pktbuf + 3, filename);
        extract_u32(pktbuf + 3 + textlen, &last_status);
    }
    return errcode;
}

// File system
int dvr_get_filecount(DVR_STORAGE_DEVID dev, u32 *count)
{
    // Build command packet
    int errcode;
    u8 pktbuf[100]; // send/recv packet buf
    u8 *pbuf = &pktbuf[1];                  // reserve space for cobs header
    int textlen;
                                            // build command:
    *pbuf++ = MSGTYPE_COMMAND;              //   message type
    *pbuf++ = CMD_GET_FILECOUNT;            //   opcode
    *pbuf++ = (u8)dev;

    // Send command and receive reply
    errcode = dvr_transact(pktbuf, sizeof(pktbuf), pbuf - &pktbuf[1]);
    if (errcode == DVR_ERR_OK) {
        extract_u32(pktbuf + 3, count);
        extract_u32(pktbuf + 7, &last_status);
    }
    return errcode;
}
        
int dvr_get_filename(DVR_STORAGE_DEVID dev, u32 index, char *filename)
{
    // Build command packet
    int errcode;
    u8 pktbuf[100]; // send/recv packet buf
    u8 *pbuf = &pktbuf[1];                  // reserve space for cobs header
    int textlen;
                                            // build command:
    *pbuf++ = MSGTYPE_COMMAND;              //   message type
    *pbuf++ = CMD_GET_FILENAME;             //   opcode
    *pbuf++ = (u8)dev;
    *pbuf++ = (u8)index;
    *pbuf++ = (u8)(index>>8);
    *pbuf++ = (u8)(index>>16);
    *pbuf++ = (u8)(index>>24);

    // Send command and receive reply
    errcode = dvr_transact(pktbuf, sizeof(pktbuf), pbuf - &pktbuf[1]);
    if (errcode == DVR_ERR_OK) {
        textlen = extract_string(pktbuf + 3, filename);
        extract_u32(pktbuf + 3 + textlen, &last_status);
    }
    return errcode;
}
        
int dvr_delete_file(DVR_STORAGE_DEVID dev, const char *filename)
{
    // Build command packet
    int errcode;
    u8 pktbuf[100]; // send/recv packet buf
    u8 *pbuf = &pktbuf[1];                  // reserve space for cobs header
    int textlen = strlen(filename) + SIZEOF_STRING_TERMINATOR;
                                            // build command:
    *pbuf++ = MSGTYPE_COMMAND;              //   message type
    *pbuf++ = CMD_DELETE_FILE;              //   opcode
    *pbuf++ = (u8)dev;
    memcpy(pbuf, filename, textlen);
    pbuf += textlen;

    // Send command and receive reply
    errcode = dvr_transact(pktbuf, sizeof(pktbuf), pbuf - &pktbuf[1]);
    if (errcode == DVR_ERR_OK) {
        extract_u32(pktbuf + 3, &last_status);
    }
    return errcode;
}

        
int dvr_rename_file(DVR_STORAGE_DEVID dev, const char *oldname, const char *newname)
{
    // Build command packet
    int errcode;
    u8 pktbuf[100]; // send/recv packet buf
    u8 *pbuf = &pktbuf[1];                  // reserve space for cobs header
    int textlen;
                                            // build command:
    *pbuf++ = MSGTYPE_COMMAND;              //   message type
    *pbuf++ = CMD_RENAME_FILE;              //   opcode
    *pbuf++ = (u8)dev;
    textlen = strlen(oldname) + SIZEOF_STRING_TERMINATOR;
    memcpy(pbuf, oldname, textlen);
    pbuf += textlen;
    textlen = strlen(newname) + SIZEOF_STRING_TERMINATOR;
    memcpy(pbuf, newname, textlen);
    pbuf += textlen;

    // Send command and receive reply
    errcode = dvr_transact(pktbuf, sizeof(pktbuf), pbuf - &pktbuf[1]);
    if (errcode == DVR_ERR_OK) {
        extract_u32(pktbuf + 3, &last_status);
    }
    return errcode;
}






// Set video overlay text and get dvr status
int dvr_set_osd_text(u8 region, char *text)
{
    // Build command packet
    int errcode;
    u8 pktbuf[100]; // send/recv packet buf
    u8 *pbuf = &pktbuf[1];                  // reserve space for cobs header
    int textlen = strlen(text) + SIZEOF_STRING_TERMINATOR;    // length of overlay text includes end-of-string marker
                                            // build command:
    *pbuf++ = MSGTYPE_COMMAND;              //   message type
    *pbuf++ = CMD_SET_OSD_TEXT;             //   opcode
    *pbuf++ = region;                       //   osd region
    memcpy(pbuf, text, textlen);            //   osd text
    pbuf += textlen;

    // Send command and receive reply
    errcode = dvr_transact(pktbuf, sizeof(pktbuf), pbuf - &pktbuf[1]);
    if (errcode == DVR_ERR_OK)
        extract_u32(pktbuf + 3, &last_status);
    return errcode;
}

// Position overlay on the video and get dvr status
int dvr_set_osd_position(u8 region, u16 x, u16 y)
{
    // Build command packet
    int errcode;
    u8 pktbuf[100];
    u8 *pbuf = &pktbuf[1];                  // reserve space for cobs header
                                            // build command:
    *pbuf++ = MSGTYPE_COMMAND;              //   message type
    *pbuf++ = CMD_SET_OSD_POSITION;         //   opcode
    *pbuf++ = region;                       //   osd region
    *pbuf++ = (u8)y;                        //   osd coordinates
    *pbuf++ = (u8)(y >> 8);
    *pbuf++ = (u8)x;
    *pbuf++ = (u8)(x >> 8);

    // Send command and receive reply
    errcode = dvr_transact(pktbuf, sizeof(pktbuf), pbuf - &pktbuf[1]);
    if (errcode == DVR_ERR_OK)
        extract_u32(pktbuf + 3, &last_status);
    return errcode;
}

// Enable the overlay on the video by setting the font size 0=disable 1=8x13 2=16x16
int dvr_set_osd_enable(u8 region, u8 enable)
{
    // Build command packet
    int errcode;
    u8 pktbuf[100];
    u8 *pbuf = &pktbuf[1];                  // reserve space for cobs header
                                            // build command:
    *pbuf++ = MSGTYPE_COMMAND;              //   message type
    *pbuf++ = CMD_SET_OSD_ENABLE;           //   opcode
    *pbuf++ = region;                       //   osd region
    *pbuf++ = enable;                       //   osd enable

    // Send command and receive reply
    errcode = dvr_transact(pktbuf, sizeof(pktbuf), pbuf - &pktbuf[1]);
    if (errcode == DVR_ERR_OK)
        extract_u32(pktbuf + 3, &last_status);
    return errcode;
}

int dvr_set_osd_background(u8 region, DVR_OSD_BGMODE bgmode)
{
    // Build command packet
    int errcode;
    u8 pktbuf[100];
    u8 *pbuf = &pktbuf[1];                  // reserve space for cobs header
                                            // build command:
    *pbuf++ = MSGTYPE_COMMAND;              //   message type
    *pbuf++ = CMD_SET_OSD_BACKGROUND;       //   opcode
    *pbuf++ = region;                       //   osd region
    *pbuf++ = bgmode;

    // Send command and receive reply
    errcode = dvr_transact(pktbuf, sizeof(pktbuf), pbuf - &pktbuf[1]);
    if (errcode == DVR_ERR_OK)
        extract_u32(pktbuf + 3, &last_status);
    return errcode;
}

int dvr_set_osd_clock_format(DVR_DATE_FORMAT datefmt, DVR_TIME_FORMAT timefmt)
{
    // Build command packet
    int errcode;
    u8 pktbuf[100];
    u8 *pbuf = &pktbuf[1];                  // reserve space for cobs header
                                            // build command:
    *pbuf++ = MSGTYPE_COMMAND;              //   message type
    *pbuf++ = CMD_SET_OSD_CLOCK_FORMAT;     //   opcode
    *pbuf++ = datefmt;
    *pbuf++ = timefmt;

    // Send command and receive reply
    errcode = dvr_transact(pktbuf, sizeof(pktbuf), pbuf - &pktbuf[1]);
    if (errcode == DVR_ERR_OK)
        extract_u32(pktbuf + 3, &last_status);
    return errcode;
}

int dvr_get_osd_clock_format(DVR_DATE_FORMAT *datefmt, DVR_TIME_FORMAT *timefmt)
{
    // Build command packet
    int errcode;
    u8 pktbuf[100];
    u8 *pbuf = &pktbuf[1];                  // reserve space for cobs header
                                            // build command:
    *pbuf++ = MSGTYPE_COMMAND;              //   message type
    *pbuf++ = CMD_GET_OSD_CLOCK_FORMAT;     //   opcode

    // Send command and receive reply
    errcode = dvr_transact(pktbuf, sizeof(pktbuf), pbuf - &pktbuf[1]);
    if (errcode == DVR_ERR_OK) {
        *datefmt = (DVR_DATE_FORMAT)pktbuf[3];
        *timefmt = (DVR_DATE_FORMAT)pktbuf[4];
        extract_u32(pktbuf + 5, &last_status);
    }
    return errcode;
}
        
int dvr_get_osd_settings(u8 region, u8 *enable, DVR_OSD_BGMODE *bgmode)
{
    // Build command packet
    int errcode;
    u8 pktbuf[100];
    u8 *pbuf = &pktbuf[1];                  // reserve space for cobs header
                                            // build command:
    *pbuf++ = MSGTYPE_COMMAND;              //   message type
    *pbuf++ = CMD_GET_OSD_SETTINGS;         //   opcode
    *pbuf++ = region;                       //   osd region

    // Send command and receive reply
    errcode = dvr_transact(pktbuf, sizeof(pktbuf), pbuf - &pktbuf[1]);
    if (errcode == DVR_ERR_OK) {
        *enable = pktbuf[3];
        *bgmode = (DVR_OSD_BGMODE)pktbuf[4];
        extract_u32(pktbuf + 5, &last_status);
    }
    return errcode;
}

int dvr_osd_image_create(u8 region, u16 len)
{
    // Build command packet
    int errcode;
    u8 pktbuf[100];
    u8 *pbuf = &pktbuf[1];                  // reserve space for cobs header
                                            // build command:
    *pbuf++ = MSGTYPE_COMMAND;              //   message type
    *pbuf++ = CMD_OSD_IMAGE_CREATE;         //   opcode
    *pbuf++ = region;                       //   osd region
    *pbuf++ = (u8)len;
    *pbuf++ = (u8)(len >> 8);

    // Send command and receive reply
    errcode = dvr_transact(pktbuf, sizeof(pktbuf), pbuf - &pktbuf[1]);
    if (errcode == DVR_ERR_OK) {
        extract_u32(pktbuf + 3, &last_status);
    }
    return errcode;
}
        
int dvr_osd_image_upload_chunk(u8 region, u16 addr, u16 len, u8 *payload)
{
    // Build command packet
    int errcode;
    u8 pktbuf[100];
    u8 *pbuf = &pktbuf[1];                  // reserve space for cobs header
                                            // build command:
    *pbuf++ = MSGTYPE_COMMAND;              //   message type
    *pbuf++ = CMD_OSD_IMAGE_UPLOAD_CHUNK;   //   opcode
    *pbuf++ = region;                       //   osd region
    *pbuf++ = (u8)addr;
    *pbuf++ = (u8)(addr >> 8);
    *pbuf++ = (u8)len;
    *pbuf++ = (u8)(len >> 8);
    memcpy(pbuf, payload, len);
    pbuf += len;

    // Send command and receive reply
    errcode = dvr_transact(pktbuf, sizeof(pktbuf), pbuf - &pktbuf[1]);
    if (errcode == DVR_ERR_OK) {
        extract_u32(pktbuf + 3, &last_status);
    }
    return errcode;
}
        
int dvr_osd_image_install(u8 region, DVR_OSD_IMAGETYPE imagetype)
{
    // Build command packet
    int errcode;
    u8 pktbuf[100];
    u8 *pbuf = &pktbuf[1];                  // reserve space for cobs header
                                            // build command:
    *pbuf++ = MSGTYPE_COMMAND;              //   message type
    *pbuf++ = CMD_OSD_IMAGE_INSTALL;        //   opcode
    *pbuf++ = region;                       //   osd region
    *pbuf++ = (u8)imagetype;

    // Send command and receive reply
    errcode = dvr_transact(pktbuf, sizeof(pktbuf), pbuf - &pktbuf[1]);
    if (errcode == DVR_ERR_OK) {
        extract_u32(pktbuf + 3, &last_status);
    }
    return errcode;
}

// Video adjust
int dvr_set_video_adjust(DVR_VIDEO_ADJ_ID param_id, u8 value)
{
    // Build command packet
    int errcode;
    u8 pktbuf[100];
    u8 *pbuf = &pktbuf[1];                  // reserve space for cobs header
                                            // build command:
    *pbuf++ = MSGTYPE_COMMAND;              //   message type
    *pbuf++ = CMD_SET_VIDEO_ADJUST;         //   opcode
    *pbuf++ = param_id;                     //   parameter
    *pbuf++ = value;

    // Send command and receive reply
    errcode = dvr_transact(pktbuf, sizeof(pktbuf), pbuf - &pktbuf[1]);
    if (errcode == DVR_ERR_OK) {
        extract_u32(pktbuf + 3, &last_status);
    }
    return errcode;
}

int dvr_get_video_adjust(DVR_VIDEO_ADJ_ID param_id, u8 *value)
{
    // Build command packet
    int errcode;
    u8 pktbuf[100];
    u8 *pbuf = &pktbuf[1];                  // reserve space for cobs header
                                            // build command:
    *pbuf++ = MSGTYPE_COMMAND;              //   message type
    *pbuf++ = CMD_GET_VIDEO_ADJUST;         //   opcode
    *pbuf++ = param_id;                     //   parameter

    // Send command and receive reply
    errcode = dvr_transact(pktbuf, sizeof(pktbuf), pbuf - &pktbuf[1]);
    if (errcode == DVR_ERR_OK) {
        *value = pktbuf[3];
        extract_u32(pktbuf + 4, &last_status);
    }
    return errcode;
}


// Video configuration
int dvr_set_video_config(u8 format, u8 deint, u8 decimate)
{
    // Build command packet
    int errcode;
    u8 pktbuf[100];
    u8 *pbuf = &pktbuf[1];                  // reserve space for cobs header
                                            // build command:
    *pbuf++ = MSGTYPE_COMMAND;              //   message type
    *pbuf++ = CMD_SET_VIDEO_CFG;            //   opcode
    *pbuf++ = format;
    *pbuf++ = deint;
    *pbuf++ = decimate;

    // Send command and receive reply
    errcode = dvr_transact(pktbuf, sizeof(pktbuf), pbuf - &pktbuf[1]);
    if (errcode == DVR_ERR_OK) {
        extract_u32(pktbuf + 3, &last_status);
    }
    return errcode;
}

int dvr_get_video_config(u8 *format, u8 *deint, u8 *decimate)
{
    // Build command packet
    int errcode;
    u8 pktbuf[100];
    u8 *pbuf = &pktbuf[1];                  // reserve space for cobs header
                                            // build command:
    *pbuf++ = MSGTYPE_COMMAND;              //   message type
    *pbuf++ = CMD_GET_VIDEO_CFG;            //   opcode

    // Send command and receive reply
    errcode = dvr_transact(pktbuf, sizeof(pktbuf), pbuf - &pktbuf[1]);
    if (errcode == DVR_ERR_OK) {
        *format = pktbuf[3];
        *deint = pktbuf[4];
        *decimate = pktbuf[5];
        extract_u32(pktbuf + 6, &last_status);
    }
    return errcode;
}
        

// Audio
int dvr_set_audio_input(DVR_AUDIO_INPUT src)
{
    // Build command packet
    int errcode;
    u8 pktbuf[100];
    u8 *pbuf = &pktbuf[1];                  // reserve space for cobs header
                                            // build command:
    *pbuf++ = MSGTYPE_COMMAND;              //   message type
    *pbuf++ = CMD_SET_AUDIO_INPUT;          //   opcode
    *pbuf++ = (u8)src;

    // Send command and receive reply
    errcode = dvr_transact(pktbuf, sizeof(pktbuf), pbuf - &pktbuf[1]);
    if (errcode == DVR_ERR_OK) {
        extract_u32(pktbuf + 3, &last_status);
    }
    return errcode;
}
        
int dvr_set_volume(DVR_AUDIO_CHAN chansel, u8 volume)
{
    // Build command packet
    int errcode;
    u8 pktbuf[100];
    u8 *pbuf = &pktbuf[1];                  // reserve space for cobs header
                                            // build command:
    *pbuf++ = MSGTYPE_COMMAND;              //   message type
    *pbuf++ = CMD_SET_VOLUME;               //   opcode
    *pbuf++ = (u8)chansel;
    *pbuf++ = volume;

    // Send command and receive reply
    errcode = dvr_transact(pktbuf, sizeof(pktbuf), pbuf - &pktbuf[1]);
    if (errcode == DVR_ERR_OK) {
        extract_u32(pktbuf + 3, &last_status);
    }
    return errcode;
}
        
int dvr_set_mute(DVR_AUDIO_CHAN chansel, u8 mute)
{
    // Build command packet
    int errcode;
    u8 pktbuf[100];
    u8 *pbuf = &pktbuf[1];                  // reserve space for cobs header
                                            // build command:
    *pbuf++ = MSGTYPE_COMMAND;              //   message type
    *pbuf++ = CMD_SET_MUTE;                 //   opcode
    *pbuf++ = (u8)chansel;
    *pbuf++ = mute;

    // Send command and receive reply
    errcode = dvr_transact(pktbuf, sizeof(pktbuf), pbuf - &pktbuf[1]);
    if (errcode == DVR_ERR_OK) {
        extract_u32(pktbuf + 3, &last_status);
    }
    return errcode;
}
        
int dvr_set_autovol(DVR_AUDIO_CHAN chansel, u8 agc)
{
    // Build command packet
    int errcode;
    u8 pktbuf[100];
    u8 *pbuf = &pktbuf[1];                  // reserve space for cobs header
                                            // build command:
    *pbuf++ = MSGTYPE_COMMAND;              //   message type
    *pbuf++ = CMD_SET_AUTOVOL;              //   opcode
    *pbuf++ = (u8)chansel;
    *pbuf++ = agc;

    // Send command and receive reply
    errcode = dvr_transact(pktbuf, sizeof(pktbuf), pbuf - &pktbuf[1]);
    if (errcode == DVR_ERR_OK) {
        extract_u32(pktbuf + 3, &last_status);
    }
    return errcode;
}
        
int dvr_get_audio_settings(DVR_AUDIO_INPUT *src, u8 *Lvol, u8 *Rvol, u8 *Lmute, 
		u8 *Rmute, u8 *Lagc, u8 *Ragc)
{
    // Build command packet
    int errcode;
    u8 pktbuf[100];
    u8 *pbuf = &pktbuf[1];                  // reserve space for cobs header
                                            // build command:
    *pbuf++ = MSGTYPE_COMMAND;              //   message type
    *pbuf++ = CMD_GET_AUDIO_SETTINGS;       //   opcode

    // Send command and receive reply
    errcode = dvr_transact(pktbuf, sizeof(pktbuf), pbuf - &pktbuf[1]);
    if (errcode == DVR_ERR_OK) {
        *src = (DVR_AUDIO_INPUT)pktbuf[3];
        *Lvol = pktbuf[4];
        *Rvol = pktbuf[5];
        *Lmute = pktbuf[6];
        *Rmute = pktbuf[7];
        *Lagc = pktbuf[8];
        *Ragc = pktbuf[9];
        extract_u32(pktbuf + 10, &last_status);
    }
    return errcode;
}

// Incremental encoder
int dvr_set_encoder_counts(u8 chan, u32 count)
{
    // Build command packet
    int errcode;
    u8 pktbuf[100];
    u8 *pbuf = &pktbuf[1];                  // reserve space for cobs header
                                            // build command:
    *pbuf++ = MSGTYPE_COMMAND;              //   message type
    *pbuf++ = CMD_SET_ENCODER_COUNTS;       //   opcode
    *pbuf++ = chan;
    *pbuf++ = (u8)count;
    *pbuf++ = (u8)(count>>8);
    *pbuf++ = (u8)(count>>16);
    *pbuf++ = (u8)(count>>24);

    // Send command and receive reply
    errcode = dvr_transact(pktbuf, sizeof(pktbuf), pbuf - &pktbuf[1]);
    if (errcode == DVR_ERR_OK) {
        extract_u32(pktbuf + 3, &last_status);
    }
    return errcode;
}
        
int dvr_get_encoder_counts(u8 chan, u32 *count)
{
    // Build command packet
    int errcode;
    u8 pktbuf[100];
    u8 *pbuf = &pktbuf[1];                  // reserve space for cobs header
                                            // build command:
    *pbuf++ = MSGTYPE_COMMAND;              //   message type
    *pbuf++ = CMD_GET_ENCODER_COUNTS;       //   opcode
    *pbuf++ = chan;

    // Send command and receive reply
    errcode = dvr_transact(pktbuf, sizeof(pktbuf), pbuf - &pktbuf[1]);
    if (errcode == DVR_ERR_OK) {
        extract_u32(pktbuf + 3, count);
        extract_u32(pktbuf + 7, &last_status);
    }
    return errcode;
}
        
int dvr_set_encoder_scalar(u8 chan, double scale)
{
    // Build command packet
    int errcode;
    u8 pktbuf[100];
    u8 *pbuf = &pktbuf[1];                  // reserve space for cobs header
                                            // build command:
    *pbuf++ = MSGTYPE_COMMAND;              //   message type
    *pbuf++ = CMD_SET_ENCODER_SCALAR;       //   opcode
    *pbuf++ = chan;
    memcpy(pbuf, &scale, sizeof(double));
    pbuf += sizeof(double);

    // Send command and receive reply
    errcode = dvr_transact(pktbuf, sizeof(pktbuf), pbuf - &pktbuf[1]);
    if (errcode == DVR_ERR_OK) {
        extract_u32(pktbuf + 3, &last_status);
    }
    return errcode;
}
        
int dvr_use_offboard_encoder(u8 chan, u8 enable)
{
    // Build command packet
    int errcode;
    u8 pktbuf[100];
    u8 *pbuf = &pktbuf[1];                  // reserve space for cobs header
                                            // build command:
    *pbuf++ = MSGTYPE_COMMAND;              //   message type
    *pbuf++ = CMD_USE_OFFBOARD_ENCODER;     //   opcode
    *pbuf++ = chan;
    *pbuf++ = enable;

    // Send command and receive reply
    errcode = dvr_transact(pktbuf, sizeof(pktbuf), pbuf - &pktbuf[1]);
    if (errcode == DVR_ERR_OK) {
        extract_u32(pktbuf + 3, &last_status);
    }
    return errcode;
}
        
int dvr_get_encoder_settings(u8 chan, double *scale, u8 *use_offboard)
{
    // Build command packet
    int errcode;
    u8 pktbuf[100];
    u8 *pbuf = &pktbuf[1];                  // reserve space for cobs header
                                            // build command:
    *pbuf++ = MSGTYPE_COMMAND;              //   message type
    *pbuf++ = CMD_GET_ENCODER_SETTINGS;     //   opcode
    *pbuf++ = chan;

    // Send command and receive reply
    errcode = dvr_transact(pktbuf, sizeof(pktbuf), pbuf - &pktbuf[1]);
    if (errcode == DVR_ERR_OK) {
        extract_double(pktbuf + 3, scale);
        *use_offboard = pktbuf[11];
        extract_u32(pktbuf + 12, &last_status);
    }
    return errcode;
}


// Menu system
int dvr_menu_display(u8 enable)
{
    // Build command packet
    int errcode;
    u8 pktbuf[100];
    u8 *pbuf = &pktbuf[1];                  // reserve space for cobs header
                                            // build command:
    *pbuf++ = MSGTYPE_COMMAND;              //   message type
    *pbuf++ = CMD_MENU_DISPLAY;             //   opcode
    *pbuf++ = enable;

    // Send command and receive reply
    errcode = dvr_transact(pktbuf, sizeof(pktbuf), pbuf - &pktbuf[1]);
    if (errcode == DVR_ERR_OK) {
        extract_u32(pktbuf + 3, &last_status);
    }
    return errcode;
}

int dvr_menu_move_cursor(u8 up)
{
    // Build command packet
    int errcode;
    u8 pktbuf[100];
    u8 *pbuf = &pktbuf[1];                  // reserve space for cobs header
                                            // build command:
    *pbuf++ = MSGTYPE_COMMAND;              //   message type
    *pbuf++ = CMD_MENU_MOVE_CURSOR;         //   opcode
    *pbuf++ = up;

    // Send command and receive reply
    errcode = dvr_transact(pktbuf, sizeof(pktbuf), pbuf - &pktbuf[1]);
    if (errcode == DVR_ERR_OK) {
        extract_u32(pktbuf + 3, &last_status);
    }
    return errcode;
}

int dvr_menu_enter(u8 *menuid, u8 *executable, char *menutext)
{
    // Build command packet
    int errcode;
    u8 pktbuf[100];
    u8 *pbuf = &pktbuf[1];                  // reserve space for cobs header
    int textlen;
                                            // build command:
    *pbuf++ = MSGTYPE_COMMAND;              //   message type
    *pbuf++ = CMD_MENU_ENTER;               //   opcode

    // Send command and receive reply
    errcode = dvr_transact(pktbuf, sizeof(pktbuf), pbuf - &pktbuf[1]);
    if (errcode == DVR_ERR_OK) {
        *menuid = pktbuf[3];
        *executable = pktbuf[4];
        textlen = extract_string(pktbuf+5, menutext);
        extract_u32(pktbuf + 5 + textlen, &last_status);
    }
    return errcode;
}

int dvr_menu_exit(void)
{
    // Build command packet
    int errcode;
    u8 pktbuf[100];
    u8 *pbuf = &pktbuf[1];                  // reserve space for cobs header
                                            // build command:
    *pbuf++ = MSGTYPE_COMMAND;              //   message type
    *pbuf++ = CMD_MENU_EXIT;                //   opcode

    // Send command and receive reply
    errcode = dvr_transact(pktbuf, sizeof(pktbuf), pbuf - &pktbuf[1]);
    if (errcode == DVR_ERR_OK) {
        extract_u32(pktbuf + 3, &last_status);
    }
    return errcode;
}
