////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// File:        simple_sdvr_codec.h
// Function:    Sensoray DVR communication protocol, example utility functions and usage
// Copyright:   Copyright (c) Sensoray Co., Inc 2020
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include <time.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/time.h>
#include <unistd.h>
#include <libgen.h>

#include "dvr_simple.h"

static char *get_err_string(int err)
{
    switch(err) {
#define CASE_STR(x) case x: return #x
    CASE_STR(DVR_ERR_INVALID_OPCODE);
    CASE_STR(DVR_ERR_ILLEGAL_ARGVAL);
    CASE_STR(DVR_ERR_MISSING_ARG);
    CASE_STR(DVR_ERR_FUNC_PROHIBITED);
    CASE_STR(DVR_ERR_FILE_NOT_FOUND);
// DVR/API
    CASE_STR(DVR_ERR_MALLOC);
// API and transaction:
    CASE_STR(DVR_ERR_OPEN);
    CASE_STR(DVR_ERR_CLOSE);
    CASE_STR(DVR_ERR_SOCKCREATE);
    CASE_STR(DVR_ERR_SOCKBIND);
    CASE_STR(DVR_ERR_RXTHREAD);
    CASE_STR(DVR_ERR_TXTHREAD);
    CASE_STR(DVR_ERR_MUTEXCREATE);
    CASE_STR(DVR_ERR_LOCKCLOSED);
    CASE_STR(DVR_ERR_SOCKCLOSED);
    CASE_STR(DVR_ERR_OVERSIZE);
    CASE_STR(DVR_ERR_PACKETSEND);
    CASE_STR(DVR_ERR_PACKETREAD);
    CASE_STR(DVR_ERR_RUNTPACKET);
    CASE_STR(DVR_ERR_MISSINGDELIM);
    CASE_STR(DVR_ERR_CHECKSUM);
    CASE_STR(DVR_ERR_MSGTYPE);
    CASE_STR(DVR_ERR_DEPLETED);
    CASE_STR(DVR_ERR_STRINGBUF);
    CASE_STR(DVR_ERR_TIMEOUT);
    CASE_STR(DVR_ERR_EVENTTYPE);
    CASE_STR(DVR_ERR_STORAGE);
    CASE_STR(DVR_ERR_VIDLOCK);
// Menu system
    CASE_STR(DVR_ERR_MENUPANE_NOTFOUND);
    CASE_STR(DVR_ERR_MENUITEM_NOTFOUND);
    CASE_STR(DVR_ERR_MENUALREADYOPEN);
    CASE_STR(DVR_ERR_MENUCLOSED);
    }
    return "";
}

static char *get_status_string(void)
{
    static char s[200];
    u32 st = dvr_get_last_status();
    
    strcpy(s, "");
    if (st & DVR_STAT_RECORDING) strcat(s, "REC ");
    if (st & DVR_STAT_PLAYING) strcat(s, "PLAY ");
    if (st & DVR_STAT_VIDEO_LOCK) strcat(s, "VIDLOCK ");
    if (st & DVR_STAT_KBD_ATTACHED) strcat(s, "KBD ");
    if (st & DVR_STAT_MEM_ATTACHED) strcat(s, "MEM ");
    if (st & DVR_STAT_WIFI_ENABLED) strcat(s, "WIFI ");
    if (st & DVR_STAT_RECORD_PAUSED) strcat(s, "RECPAUSE ");
    if (st & DVR_STAT_PLAY_PAUSED) strcat(s, "PLAYPAUSE ");
    if (st & DVR_STAT_MENU_OPEN) strcat(s, "MENU ");
    if (st & DVR_STAT_VIEW_SNAPSHOT) strcat(s, "SNAPSHOT ");
    if (st & DVR_STAT_WRITE_ERROR) strcat(s, "ERRWRITE ");
    
    return s;
}

#define TEST(x)  \
do {  \
	int errcode = (x); \
	if (errcode != DVR_ERR_OK) { \
		fprintf(stderr, "%s: err=%d (%s) [%s]\n", #x, errcode, get_err_string(errcode), get_status_string()); \
		return errcode; \
	} \
} while (0)


////////////////////////////////////////////////////////////////////////////
// Example usage: Send overlay text to DVR using in-place encode/decode.

int main(int unused)
{
    int errcode;
    u32 status;     // dvr status flags
    
    TEST( open_com_port() );
    

    // General ------------------------------------------
    {
	u32 version;
        TEST( dvr_get_api_version(&version) );
        printf("The DVR API version is %d.%d.%d\n",
                (version >> 24) & 0xff,
                (version >> 16) & 0xff,
                (version & 0xffff));
        if (version < 0x01010000) {
            printf("DVR API version needs to be at least 1.1\n");
            return 0;
        }
    }
    
    {
        u32 hw_version;
        char model[100];
        TEST( dvr_get_hardware_info(&hw_version, model) );
        printf("The hardware info is %d.%d.%d, model=%s\n", 
                (hw_version >> 24) & 0xff,
                (hw_version >> 16) & 0xff,
                (hw_version & 0xffff),
                model);
    }
    
    {
        u32 status;
        TEST( dvr_get_dvr_status(&status) );
        printf("The DVR status is 0x%x (%s%s%s%s%s%s%s)\n", status,
                status & DVR_STAT_RECORDING ? "recording " : "",
                status & DVR_STAT_PLAYING ? "playing " : "",
                status & DVR_STAT_VIDEO_LOCK ? "videolock " : "",
                status & DVR_STAT_KBD_ATTACHED ? "keyboard " : "",
                status & DVR_STAT_MEM_ATTACHED ? "mem " : "",
                status & DVR_STAT_WIFI_ENABLED ? "wifi " : "",
                status & DVR_STAT_MENU_OPEN ? "menu " : ""
                );
    }
    
    {
        DVR_DISPLAY_MODE disp_mode;
        TEST( dvr_get_general_settings(&disp_mode) );
        printf("The general settings is 0x%x\n", disp_mode);

        // FIXME this is not working
        TEST( dvr_set_display_mode(DISPMODE_COLORBAR) );
        printf("Setting colorbars\n");
        sleep(1);
        
        TEST( dvr_set_display_mode(disp_mode) );
        printf("Restoring original display mode\n");
    }
    
    
    // Clock ------------------------------------------
    {
	time_t t;
        u32 clock;
        TEST( dvr_get_datetime(&clock) );
        u32 orig = clock;
        
        t = clock;
        printf("The clock on the DVR is %s\n", asctime(gmtime(&t)));
	
	struct timeval tv;
	struct timezone tz;
	gettimeofday(&tv, &tz);
	t = tv.tv_sec;
	clock = timegm(localtime(&t));  // adjust for local timezone
	
        printf("Set clock on the DVR to %s\n", asctime(localtime(&t)));

	TEST( dvr_set_datetime(clock) );
        
        TEST( dvr_get_datetime(&clock) );
        
        t = clock;
        printf("The clock on the DVR is %s\n", asctime(gmtime(&t)));
        
        TEST( dvr_set_datetime(orig) );
        printf("Restoring original clock\n");
    }
    
    // Recording (requires internal storage)
    char recorded_filename[MAX_FILENAME_SIZE];
    if (0) {
        char filename[MAX_FILENAME_SIZE];

        TEST( dvr_record_stop() != DVR_ERR_FUNC_PROHIBITED);
        TEST( dvr_record_start(BOTH, RECFMT_MP4, 0 /*default*/, filename) );
        printf("recording started, filename=%s, status: %s\n", filename, get_status_string());
        strcpy(recorded_filename, basename(filename));

        sleep(3);
        TEST( dvr_record_pause() );
        printf("paused, status: %s\n", get_status_string());
        
        sleep(2);
    
        TEST( dvr_record_resume() );
        printf("resumed, status: %s\n", get_status_string());
        
        sleep(3);
        TEST( dvr_record_stop() );
        printf("stopped, status: %s\n", get_status_string());
    }

    // Playback
    if (0) {
        TEST( dvr_play_start(INTERNAL, recorded_filename) );
        printf("playback started, filename=%s, status: %s\n", recorded_filename, get_status_string());
        sleep(2);
        
        TEST( dvr_play_pause() );
        printf("paused, status: %s\n", get_status_string());
        
        sleep(2);
    
        TEST( dvr_play_resume() );
        printf("resumed, status: %s\n", get_status_string());
        
        sleep(2);
        TEST( dvr_play_stop() );
        printf("stopped, status: %s\n", get_status_string());
        
    }
    if (0) {
        TEST( dvr_play_start(EXTERNAL, recorded_filename) );
        printf("playback started, filename=%s, status: %s\n", recorded_filename, get_status_string());
        sleep(2);
        
        TEST( dvr_play_pause() );
        printf("paused, status: %s\n", get_status_string());
        
        sleep(2);
    
        TEST( dvr_play_resume() );
        printf("resumed, status: %s\n", get_status_string());
        
        sleep(2);
        TEST( dvr_play_stop() );
        printf("stopped, status: %s\n", get_status_string());
        
    }
    
    // Snapshot (internal)
    {
        char filename[MAX_FILENAME_SIZE];
        TEST( dvr_grab_snapshot(BOTH, 90, filename) );
        printf("snapshot: filename=%s\n", filename);
        
        TEST( dvr_view_snapshot(EXTERNAL, filename) );
        printf("viewing snapshot, status: %s\n", get_status_string());
        sleep(5);
        TEST( dvr_snapshot_stop() );
        printf("stop snapshot, status: %s\n", get_status_string());
        
        TEST( dvr_rename_file(INTERNAL, filename, "trash.jpg") );
        printf("deleting snapshot\n");
        TEST( dvr_delete_file(INTERNAL, "trash.jpg") );
    }
    
    // File system
    {
        u32 i, count, devid = INTERNAL;
        
        TEST( dvr_get_filecount(devid, &count) );
        printf("internal filecount: %d\n", count);
        for (i = 0; i < count; i++) {
            char filename[MAX_FILENAME_SIZE];
            TEST( dvr_get_filename(devid, i, filename) );
            printf("file %d: %s\n", i, filename);
        }
    }
    {
        u32 i, count, devid = EXTERNAL;
        
        TEST( dvr_get_filecount(devid, &count) );
        printf("internal filecount: %d\n", count);
        for (i = 0; i < count; i++) {
            char filename[MAX_FILENAME_SIZE];
            TEST( dvr_get_filename(devid, i, filename) );
            printf("file %d: %s\n", i, filename);
        }
    }
    
    // Video adjust
    {
        u8 brightness, autobright, contrast, hue, saturation, chroma_filt, luma_filt, comb_filt;
        TEST( dvr_get_video_adjust(ADJ_BRIGHTNESS, &brightness) );
        TEST( dvr_get_video_adjust(ADJ_AUTOBRIGHT, &autobright) );
        TEST( dvr_get_video_adjust(ADJ_CONTRAST, &contrast) );
        TEST( dvr_get_video_adjust(ADJ_HUE, &hue) );
        TEST( dvr_get_video_adjust(ADJ_SATURATION, &saturation) );
        TEST( dvr_get_video_adjust(ADJ_CHROMA_FILT, &chroma_filt) );
        TEST( dvr_get_video_adjust(ADJ_LUMA_FILT, &luma_filt) );
        TEST( dvr_get_video_adjust(ADJ_COMB_FILT, &comb_filt) );
        
        printf("Video adjustments:\n"
                "\tbrightness=%d\n"
                "\tautobright=%d\n"
                "\tcontrast=%d\n"
                "\thue=%d\n"
                "\tsaturation=%d\n"
                "\tchroma_filt=%d\n"
                "\tluma_filt=%d\n"
                "\tcomb_filt=%d\n",
                brightness, autobright, contrast, hue, saturation, chroma_filt, luma_filt, comb_filt);
                

        printf("Restoring original adjustments\n");
        TEST( dvr_set_video_adjust(ADJ_BRIGHTNESS, brightness) );
        TEST( dvr_set_video_adjust(ADJ_AUTOBRIGHT, autobright) );
        TEST( dvr_set_video_adjust(ADJ_CONTRAST, contrast) );
        TEST( dvr_set_video_adjust(ADJ_HUE, hue) );
        TEST( dvr_set_video_adjust(ADJ_SATURATION, saturation) );
        TEST( dvr_set_video_adjust(ADJ_CHROMA_FILT, chroma_filt) );
        TEST( dvr_set_video_adjust(ADJ_LUMA_FILT, luma_filt) );
        TEST( dvr_set_video_adjust(ADJ_COMB_FILT, comb_filt) );
    }
    
    // Video configuration
    {
        u8 format, deint, decimate;
        TEST( dvr_get_video_config(&format, &deint, &decimate) );
        printf("Video config: format=%d deint=%d decimate=%d\n",
                format, deint, decimate);
        
        printf("Restoring original config\n");
        TEST( dvr_set_video_config(format, deint, decimate) );
    }
    
    // Audio
    {
        DVR_AUDIO_INPUT src;
        u8 lvol, rvol, lmute, rmute, lagc, ragc;
        
        TEST( dvr_get_audio_settings(&src, &lvol, &rvol, &lmute, &rmute, &lagc, &ragc) );
        printf("current audio settings src=%d vol=%d,%d mute=%d,%d agc=%d,%d\n",
                src, lvol, rvol, lmute, rmute, lagc, ragc);

        TEST( dvr_set_audio_input(AUDIOIN_LINE) );
        TEST( dvr_set_volume(AUDIOCHAN_LEFT, 100) );
        TEST( dvr_set_volume(AUDIOCHAN_RIGHT, 50) );
        TEST( dvr_set_mute(AUDIOCHAN_LEFT, 1) );
        TEST( dvr_set_mute(AUDIOCHAN_RIGHT, 0) );
        TEST( dvr_set_autovol(AUDIOCHAN_LEFT, 0) );
        TEST( dvr_set_autovol(AUDIOCHAN_RIGHT, 1) );
        {
            DVR_AUDIO_INPUT src;
            u8 lvol, rvol, lmute, rmute, lagc, ragc;
            TEST( dvr_get_audio_settings(&src, &lvol, &rvol, &lmute, &rmute, &lagc, &ragc) );
            printf("new audio settings src=%d vol=%d,%d mute=%d,%d agc=%d,%d\n",
                    src, lvol, rvol, lmute, rmute, lagc, ragc);
            sleep(1);
        }
        
        printf("restoring audio settings\n");
        TEST( dvr_set_audio_input(src) );
        TEST( dvr_set_volume(AUDIOCHAN_LEFT, lvol) );
        TEST( dvr_set_volume(AUDIOCHAN_RIGHT, rvol) );
        TEST( dvr_set_mute(AUDIOCHAN_LEFT, lmute) );
        TEST( dvr_set_mute(AUDIOCHAN_RIGHT, rmute) );
        TEST( dvr_set_autovol(AUDIOCHAN_LEFT, lagc) );
        TEST( dvr_set_autovol(AUDIOCHAN_RIGHT, ragc) );
        
        TEST( dvr_get_audio_settings(&src, &lvol, &rvol, &lmute, &rmute, &lagc, &ragc) );
        printf("restored audio settings src=%d vol=%d,%d mute=%d,%d agc=%d,%d\n",
                src, lvol, rvol, lmute, rmute, lagc, ragc);
    }
    
    // Incremental encoder
    {
        u32 count[2];
        double scale[2];
        u8 use_offboard[2];
        int i;
        
        for (i = 0; i <= 1; i++) {
            TEST( dvr_get_encoder_counts(i, &count[i]) );
            TEST( dvr_get_encoder_settings(i, &scale[i], &use_offboard[i]) );
            printf("encoder %d counts %u scale %f offboard %d\n", i, count[i], scale[i], use_offboard[i]);
        }
        
        TEST( dvr_set_encoder_counts(0, 314159) );
        TEST( dvr_set_encoder_counts(1, 123456) );
        TEST( dvr_set_encoder_scalar(0, 3.14159) );
        TEST( dvr_set_encoder_scalar(1, 0.12345) );
        printf("changed encoder counts and scale\n");
        sleep(5);
        {
            u32 count[2];
            double scale[2];
            u8 use_offboard[2];
            
            for (i = 0; i <= 1; i++) {
                TEST( dvr_get_encoder_counts(i, &count[i]) );
                TEST( dvr_get_encoder_settings(i, &scale[i], &use_offboard[i]) );
                printf("encoder %d counts %u scale %f offboard %d\n", i, count[i], scale[i], use_offboard[i]);
            }
        }
        
        for (i = 0; i <= 1; i++) {
            TEST( dvr_set_encoder_counts(i, count[i]) );
            TEST( dvr_set_encoder_scalar(i, scale[i]) );
        }
        printf("Restoring original encoder counts\n");
        
        
    }
    

    // OSD
    if (1) {
        int i;
        printf("Testing OSD\n");
        for (i = 0; i < 16; i++) {
            char txt[100];
            
            sprintf(txt, "OSD %d\n", i);
                    
            TEST( dvr_set_osd_text(i, txt) );

            TEST( dvr_set_osd_position(i, 50+i*20, 80+i*20) ); 

            TEST( dvr_set_osd_background(i, (i&7) < 4 ? BGMODE_BLACK : BGMODE_TRANSPARENT) );

            TEST( dvr_set_osd_enable(i, i < 8 ? 2 : 1) );
        }
        sleep(2);
        printf("Clearing OSD\n");
        for (i = 0; i < 16; i++) {
            TEST( dvr_set_osd_enable(i, 0) );
        }        
    }
    
    close_com_port();
}
